# -*- coding: utf-8 -*-
from .matproj import StaticEnergy__Vasp__Matproj
from .matproj_hse import StaticEnergy__Vasp__MatprojHse
from .matproj_hsesol import StaticEnergy__Vasp__MatprojHsesol
from .matproj_pbesol import StaticEnergy__Vasp__MatprojPbesol
from .matproj_scan import StaticEnergy__Vasp__MatprojScan
from .mit import StaticEnergy__Vasp__Mit
from .mvl_neb_endpoint import StaticEnergy__Vasp__MvlNebEndpoint
from .quality04 import StaticEnergy__Vasp__Quality04



# -*- coding: utf-8 -*-

from simmate.apps.vasp.workflows.electronic_structure.base_band_structure import (
    VaspBandStructure,
)
from simmate.apps.vasp.workflows.relaxation.matproj_hse import (
    Relaxation__Vasp__MatprojHse,
)


class ElectronicStructure__Vasp__MatprojBandStructureHse(
    VaspBandStructure, Relaxation__Vasp__MatprojHse
):
    """
    This task is a reimplementation of pymatgen's
    [MPHSERelaxSet](https://pymatgen.org/pymatgen.io.vasp.sets.html#pymatgen.io.vasp.sets.MPHSERelaxSet)
    with mode="line".

    Calculates the band structure using Materials Project HSE settings.
    """

    incar = Relaxation__Vasp__MatprojHse.incar.copy()
    incar.update(
        NSW=0,
        ISMEAR=0,
        SIGMA=0.05,
        ISYM=3,
        LCHARG=False,
        NELMIN=5,
    )



# -*- coding: utf-8 -*-

import logging
import os
import sys
from pathlib import Path

import requests

import simmate


def get_conda_env() -> str:
    """
    Grab the name of the activate conda environment and returns it as a string.
    If there is no env, then an empty string is returned.
    """
    # Check the list of python paths and grab the first path that has "envs" in it.
    # Assume we don't have a conda env until proven otherwise
    env_name = ""
    for path in sys.path:
        if "envs" in path:
            # split the path into individual folder names (os.sep gives / or \\)
            folders = path.split(os.sep)
            # the conda env name will the name immediately after the /envs.
            # example path is '/home/jacksund/anaconda3/envs/simmate_dev/lib/python3.10'
            # where we want the name simmate_dev here.
            env_name = folders[folders.index("envs") + 1]
            # once we have found this, we can exit the loop
            break

    return env_name


def get_latest_version() -> str:
    """
    Looks at the jacks/simmate repo and grabs the latest release version.
    """
    # Access the data via a web request
    response = requests.get(
        "https://api.github.com/repos/jacksund/simmate/releases/latest"
    )

    # load the version from the json response
    # [1:] simply removes the first letter "v" from something like "v1.2.3"
    latest_version = response.json()["tag_name"][1:]
    return latest_version


def check_if_using_latest_version(current_version=simmate.__version__):
    """
    Checks if there's a newer version by looking at the latest release on Github
    and comparing it to the currently installed version
    """
    latest_version = get_latest_version()

    if current_version != latest_version:
        logging.warning(
            "There is a new version of Simmate available. "
            f"You are currently using v{current_version} while v{latest_version} "
            "is the latest."
        )


def get_chemical_subsystems(chemical_system: str):
    """
    Given a chemical system, this returns all chemical systems that are also
    contained within it.

    For example, "Y-C" would return ["Y", "C", "C-Y"]. Note that the returned
    list has elements of a given system in alphabetical order (i.e. it gives
    "C-Y" and not "Y-C")

    #### Parameters

    - `chemical_system`:
        A chemical system of elements. Elements must be separated by dashes (-)

    #### Returns

    - `subsystems`:
        A list of chemical systems that make up the input chemical system.
    """

    # TODO: this code may be better located elsewhere. Maybe even as a method for
    # the Composition class or alternatively as a ChemicalSystem class.

    # I convert the system to a composition where the number of atoms dont
    # apply here. (e.g. "Ca-N" --> "Ca1 N1")
    from simmate.toolkit import Composition

    composition = Composition(chemical_system.replace("-", ""))

    return composition.chemical_subsystems


def chunk_list(full_list: list, chunk_size: int) -> list:
    """
    Yield successive n-sized chunks from a list.
    """
    for i in range(0, len(full_list), chunk_size):
        yield full_list[i : i + chunk_size]


def str_to_datatype(
    parameter: str,
    value: str,
    type_mappings: dict = {},
):
    """
    When given a parameter name and it's value as a string, this helper
    function will use the key (parameter) to determine how to convert the
    val string to the proper python datatype (int, float, bool, list...).
    A full mapping of parameters should be provided, but if a parameter is
    given that isn't mapped, the value will be leave left as a string.

    This is often ment to read to/from non-standard parameter file formats.
    For example, VASP's INCAR does not follow any standard (yaml, toml, json,
    etc.) so this function helps read values into python types.

    Example type_mapping using random VASP settings:
    ``` python
    type_mapping = {
        "LDAU": bool,
        "EDIFF": float,
        "NSW": int,
        "LDAUL": list[int],
        "LDAUU": list[float],
        "DIPOL": list[list[float]],
    }
    ```
    """

    # If the value is not a string, then assume we are already in the
    # correct format. Note, an incorrect format will throw an error
    # somewhere below, which may be tricky for beginners to traceback.
    if not isinstance(value, str):
        return value

    # next, try grabbing the type from mapping dictionary. If the parameter is
    # not mapped, then we assume it is a str.
    target_type = type_mappings.get(parameter, str)

    # Now that we know the target type to convert to, we can go through
    # decide how to handle converting the value

    if target_type == str:
        # assert type(value) == str
        return value

    elif target_type == int:
        # sometimes "1." was written to indicate an integer so check for
        # this and remove it if needed.
        if value[-1] == ".":
            value = value[:-1]
        # return the value integer
        return int(value)

    elif target_type == float:
        # return the value float
        return float(value)

    elif target_type == bool:
        # Python is weird where bool("FALSE") will return True... So I need
        # to convert the string to lowercase and read it to know what to
        # return here.
        if "t" in value.lower():
            return True
        elif "f" in value.lower():
            return False

    elif target_type == list[float]:
        final_list = []
        for item in value.split():
            # Sometimes, the values are given as "3*0.1 2*0.5" where the "*"
            # means to include that value that many times. For example, this
            # input would be the same as "0.1 0.1 0.1 0.5 0.5". We need to
            # account for this when parsing.
            if "*" in item:
                nsubitems, subitem = item.split("*")
                for n in range(int(nsubitems)):
                    final_list.append(float(subitem))
            else:
                final_list.append(float(item))
        return final_list

    elif target_type == list[int]:
        return [int(item) for item in value.split()]

    # These vectors are always 3x floats
    elif target_type == list[list[float]]:
        # convert a string of...
        #   "x1 y1 z1 x2 y2 z2 x3 y3 z3"
        # to...
        #   [x1,y1,z1,x2,y2,z2,x3,y3,z3] (list of floats)
        # and then to...
        #   [[x1,y1,z1],[x2,y2,z2],[x3,y3,z3]]
        value = [float(item) for item in value.split()]
        return [value[i : i + 3] for i in range(0, len(value), 3)]

    # If it is not in the common keys listed, just leave it as a string.
    else:
        logging.warning(
            "Unknown parameter mapping of {parameter}: {target_type}. "
            "Leaving as str."
        )
        return value



# -*- coding: utf-8 -*-

from ase.ga.standardmutations import RattleMutation
from ase.ga.utilities import closest_distances_generator
from pymatgen.io.ase import AseAtomsAdaptor

from simmate.toolkit import Structure
from simmate.toolkit.transformations.base import Transformation


class CoordinatePerturbation(Transformation):
    """
    This is a wrapper around the `RattleMutation` in ase.ga
    https://gitlab.com/ase/ase/-/blob/master/ase/ga/standardmutations.py
    """

    name = "from_ase.CoordinatePerturbation"
    io_scale = "one_to_one"
    ninput = 1
    allow_parallel = False

    @staticmethod
    def apply_transformation(
        structure: Structure,
        perturb_strength: float = 0.8,
        ratio_of_covalent_radii: float = 0.1,
    ) -> Structure:
        # ----------- SETUP (consider caching as class attribute) -------------
        # The closest_distances_generator is exactly the same as an
        # element-dependent distance matrix, except ASE puts this in dictionary
        # form. The function requires a list of element integers
        element_ints = [element.number for element in structure.composition]
        # the default of the ratio of covalent radii (0.1) is based on the ASE
        # tutorial of this function
        element_distance_matrix = closest_distances_generator(
            element_ints, ratio_of_covalent_radii
        )
        # ---------------------------------------------------------------------

        # first I need to convert the structures to an ASE atoms object
        structure_ase = AseAtomsAdaptor.get_atoms(structure)

        # now we can make the generator
        rattle = RattleMutation(
            blmin=element_distance_matrix,  # distance cutoff matrix
            n_top=int(
                structure.composition.num_atoms
            ),  # number of atoms to optimize. I set this to all,
            rattle_strength=perturb_strength,  # strength of rattling
            rattle_prop=1,  # propobility that atom is rattled
            test_dist_to_slab=True,
            # rng=np.random  # TODO: consider changing to normal dist
        )

        #!!! Their code suggests the use of .get_new_individual() but I think
        # .mutate() is what we'd like
        new_structure_ase = rattle.mutate(structure_ase)

        # if the mutation fails, None is return
        if not new_structure_ase:
            return False

        # if it was successful, we have a new Atoms object
        # now convert back to a pymatgen object
        new_structure = AseAtomsAdaptor.get_structure(new_structure_ase)

        return new_structure



# -*- coding: utf-8 -*-

from simmate.apps.vasp.workflows.relaxation.matproj_hse import (
    Relaxation__Vasp__MatprojHse,
)


class Relaxation__Vasp__MatprojHsesol(Relaxation__Vasp__MatprojHse):
    """
    This task is a reimplementation of pymatgen's
    [MPHSERelaxSet](https://pymatgen.org/pymatgen.io.vasp.sets.html#pymatgen.io.vasp.sets.MPHSERelaxSet).

    Runs a VASP relaxation calculation using Materials Project HSE settings with
    an additional HSEsol setting.
    """

    description_doc_short = "based on pymatgen's MPHSERelaxSet"

    incar = Relaxation__Vasp__MatprojHse.incar.copy()
    incar.update(dict(GGA="PS"))  # Tells VASP to use PBEsol instead of base PBE



# -*- coding: utf-8 -*-

import importlib

from django.contrib import admin
from django.urls import include, path

from simmate.configuration.django.settings import SIMMATE_APPS
from simmate.website.core import views


def get_app_urls():
    # First let's find our custom simmate apps that supply a 'urls' module. We
    # want to automatically make these avaialable in the web UI.

    extra_url_paths = []

    for app_name in SIMMATE_APPS:
        # This section of code is copied from...
        #   simmate.workflows.utilities.get_all_workflows
        # Consider making another util.
        config_modulename = ".".join(app_name.split(".")[:-1])
        config_name = app_name.split(".")[-1]
        config_module = importlib.import_module(config_modulename)
        config = getattr(config_module, config_name)
        app_path = config.name
        simple_name = app_path.split(".")[-1]

        # check if there is a urls module in the app
        #   stackoverflow.com/questions/14050281
        urls_found = importlib.util.find_spec(f"{app_path}.urls") is not None

        if urls_found:
            new_path = path(
                route=f"apps/{simple_name}/",
                # set the namespace so that we can easily look up app urls
                #   https://stackoverflow.com/questions/48608894
                view=include((f"{app_path}.urls", simple_name), namespace=simple_name),
                name=simple_name,
            )
            extra_url_paths.append(new_path)

    return extra_url_paths


urlpatterns = [
    #
    # This is the path to the homepage (just simmate.org)
    path(route="", view=views.home, name="home"),
    #
    # This is the built-in admin site that django provides
    path(route="admin/", view=admin.site.urls, name="admin"),
    #
    # This is profile system with login/logout
    path(
        route="accounts/",
        view=include("allauth.urls"),
        name="accounts",
    ),
    # On login success, you will be pointed to /accounts/profile by default.
    # If you want to change this defualt, then set LOGIN_REDIRECT_URL in
    # the settings.py file
    path(route="accounts/profile/", view=views.profile, name="profile"),
    # When you sign out, you are sent to LOGOUT_REDIRECT_URL (set in settings.py)
    path(route="accounts/loginstatus/", view=views.loginstatus, name="loginstatus"),
    #
    #
    path(
        route="third-parties/",
        view=include(
            ("simmate.website.third_parties.urls", "simmate.website.third_parties"),
            namespace="third_parties",
        ),
        name="third_parties",
    ),
    #
    #
    # All local calculations are stored at this endpoint
    path(
        route="workflows/",
        view=include("simmate.website.workflows.urls"),
        name="workflows",
    ),
    #
    # This app includes core functionality, such as views for crystal structures
    # in a 3D viewport.
    path(
        route="core-components/",
        view=include(
            (
                "simmate.website.core_components.urls",
                "simmate.website.core_components",
            ),
            namespace="core_components",
        ),
        name="core_components",
    ),
    #
    # And extra one-page views
    path(route="extras/", view=views.extras, name="extras"),
    path(route="contact/", view=views.contact, name="contact"),
    path(route="about/", view=views.about, name="about"),
    #
    # Custom Simmate apps (if present)
    path(route="apps/", view=views.apps, name="apps"),
    *get_app_urls(),
]



# -*- coding: utf-8 -*-

from simmate.apps.vasp.workflows.diffusion.mit import Diffusion__Vasp__NebFromImagesMit
from simmate.apps.vasp.workflows.diffusion.neb_base import NebFromEndpointWorkflow
from simmate.apps.vasp.workflows.relaxation.mvl_neb_endpoint import (
    Relaxation__Vasp__MvlNebEndpoint,
)


class Diffusion__Vasp__NebFromEndpointsMit(NebFromEndpointWorkflow):
    """
    Runs a full diffusion analysis from a start and end supercell using NEB with
    MIT settings.
    """

    endpoint_relaxation_workflow = Relaxation__Vasp__MvlNebEndpoint
    from_images_workflow = Diffusion__Vasp__NebFromImagesMit



# -*- coding: utf-8 -*-

from .files import (
    archive_old_runs,
    copy_directory,
    empty_directory,
    get_directory,
    make_archive,
    make_error_archive,
)
from .other import (
    check_if_using_latest_version,
    chunk_list,
    get_chemical_subsystems,
    get_conda_env,
    get_latest_version,
    str_to_datatype,
)



# -*- coding: utf-8 -*-

from pathlib import Path

from simmate.apps.vasp.inputs import Incar
from simmate.engine import ErrorHandler


class ElfKpar(ErrorHandler):
    is_monitor = True
    filename_to_check = "vasp.out"
    possible_error_messages = ["ELF: KPAR>1 not implemented"]

    def correct(self, directory: Path) -> str:
        # load the INCAR file to view the current settings
        incar_filename = directory / "INCAR"
        incar = Incar.from_file(incar_filename)

        # make the fix
        incar["KPAR"] = 1
        correction = "switched KPAR to 1"

        # rewrite the INCAR with new settings
        incar.to_file(incar_filename)

        return correction



# -*- coding: utf-8 -*-

from django.apps import AppConfig


class BaderConfig(AppConfig):
    name = "simmate.apps.bader"


class CleaseConfig(AppConfig):
    name = "simmate.apps.clease"


class VaspConfig(AppConfig):
    name = "simmate.apps.vasp"


class EvolutionConfig(AppConfig):
    name = "simmate.apps.evolution"
